(*------------------------------------------------------------------------------

   GITI tool (OCaml 4.06)
   Harrison Ainsworth / HXA7241 : 2017

   http://www.hxa.name/giti

   License: AGPL -- http://www.gnu.org/licenses/agpl.html

------------------------------------------------------------------------------*)




open HxaGeneral
open Basics




(*--- values ---*)



(*--- types ---*)

type t = unit



(*--- modules ---*)

module Tuning :
sig

   (* length 9 -- thin to thick: guitar string = index + 1 *)
   type t = private Pitch.note list

   val create : Annotation.t -> t

   val note : t -> Pitch.strng -> Pitch.fret -> Pitch.note

end
=
struct

   (* length 9 -- thin to thick: guitar string = index + 1 *)
   type t = Pitch.note list


   let stdTuning : t =
      (* thin to thick -- guitar string = index + 1 *)
      let open Pitch in
      let makeNote (l:letter) (a:accident) (o:int) : note =
         { letter = l ; accident = a ; octave = Int0to8.create o }
      in
      [  makeNote E Natural 4
      ;  makeNote B Natural 3
      ;  makeNote G Natural 3
      ;  makeNote D Natural 3
      ;  makeNote A Natural 2
      ;  makeNote E Natural 2
      ;  makeNote B Natural 1
      ;  makeNote F Sharp   1
      ;  makeNote C Sharp   1  ]


   let incNote (inc:int) (note:Pitch.note) : Pitch.note =

      let noteToInt (note:Pitch.note) : int =
         let noteRelToInt (noteRel:Pitch.letter * Pitch.accident) : Int0to11.t =
            let open Pitch in
            let letInt =
               match fst noteRel with
               | C -> 0 | D -> 2 | E -> 4 | F -> 5 | G -> 7 | A -> 9 | B -> 11
            and accInt =
               match snd noteRel with
               | Natural -> 0 | Sharp -> +1 | Flat -> -1
            in
            Int0to11.create (Int.modw (letInt + accInt) 12)
         in
         let open Pitch in
         ((noteRelToInt (note.letter , note.accident)) :> int)
         + ((note.octave :> int) * 12)

      and intToNote (number:int) : Pitch.note =
         let intToNoteRel (number:Int0to11.t)
            : (Pitch.letter * Pitch.accident) =
            let open Pitch in
            match (number :> int) with
            |  0 -> ( C , Natural ) |  1 -> ( C , Sharp )
            |  2 -> ( D , Natural ) |  3 -> ( D , Sharp )
            |  4 -> ( E , Natural )
            |  5 -> ( F , Natural ) |  6 -> ( F , Sharp )
            |  7 -> ( G , Natural ) |  8 -> ( G , Sharp )
            |  9 -> ( A , Natural ) | 10 -> ( A , Sharp )
            | 11 -> ( B , Natural )
            | _  -> ( C , Natural ) (* impossible, because Int0to11 *)
         in
         (* clamps to zero lowest octave *)
         let div12 , mod12     = number / 12 , Int.modw number 12 in
         let letter , accident = intToNoteRel (Int0to11.create mod12)
         and octave            = Int0to8.create div12 in
         let open Pitch in
         { letter ; accident ; octave }
      in

      intToNote ((noteToInt note) + inc)


   let create (annot:Annotation.t) : t =

      let merge (previous:t) (item:Annotation.tuning) : t =
         let open Annotation in
         match item with
         | TuningStd _ ->
            stdTuning
         | TuningShift { semi ; _ } ->
            (* clamps to zero lowest octave *)
            List.map (incNote (semi :> int)) previous
         | TuningSpec spec ->
            (* TODO?: maybe reverse spec -- should find get first or last ? *)
            let replaceWith (spec:tuningSpec list1) (i:int) (note:Pitch.note)
               : Pitch.note =
               excToDefault
                  note
                  (fun () ->
                     let finder (ts:tuningSpec) : bool =
                        (* NB: guitar strings numbering starts at 1 *)
                        i = (ts.tuningSpecString :> int) - 1
                     in
                     (List.find finder (ofList1 spec)).tuningSpecNote)
            in
            List.mapi (replaceWith spec) previous
      in

      (* extract tuning parts : tuning list *)
      let open Annotation in
      (List.filtmap
         (function
            | Tuning  t -> Some t
            | Tempo   _ -> None
            | Version _ -> None)
         annot)
      |>
      (* merge tuning parts : t *)
      (List.fold_left merge stdTuning)
      (* TODO?: maybe reverse order of merge *)


   let note (tuning:t) (strng:Pitch.strng) (fret:Pitch.fret) : Pitch.note =

      let strngToNote (tuning:t) (strng:Int1to9.t) : Pitch.note =
         (* should never raise,
            assuming Tuning.create, hence Tuning.t length, is correct *)
         excToDefault
            (* arbitrary default *)
            Pitch.{ letter = C ; accident = Sharp ; octave = Int0to8.create 1 }
            (fun () -> List.nth tuning ((strng :> int) - 1))
      in

      incNote (fret :> int) (strngToNote tuning strng)

end


(*
module Note =
sig
   type octave   = Int0to8.t
   type accident = | Natural | Sharp | Flat
   type letter   = | A | B | C | D | E | F | G

   type t        =
      { letter   :letter ;
        accident :accident ;
        octave   :octave ; }

   val create : letter -> accident -> octave -> t

   val shift : ~inc:int -> t -> t
end
=
struct
   type octave   = Int0to8.t
   type accident = | Natural | Sharp | Flat
   type letter   = | A | B | C | D | E | F | G

   type t        =
      { letter   :letter ;
        accident :accident ;
        octave   :octave ; }

   let create (letter:letter) (accident:accident) (octave:octave) : t =
      { letter ; accident ; octave }

   let shift (inc:int) (note:t) : t =
      ...
end
*)



(*--- functions ---*)

let augment (piece:Piece.t) : t =
   (* TODO *)
   ()

(*

(* all possible leaf types
   (not just actual leafs: anything that produces a string) *)
type leaf =
   | Str        of string
   | Comment    of Line.t
   | Metadata   of Piece.metadata
   | Annotation of Annotation.t
   | Indicator  of Indicator.t
   | Sound      of Sound.t

type height = int
type tuning = Tuning.t option
type pitch  = Pitch.t option
type time   = Time.t option


let augmentBlank (piece:Piece.t) : () list =
   ...


let augmentHeight (piece:Piece.t) : int list =
   ...


let augmentTuning (piece:Piece.t) : tuning list =
   ...


let augmentPitch (piece:Piece.t) : pitch list =
   ...


let augmentTime (piece:Piece.t) : time list =
   ...


let flattenWhole (piece:Piece.t) : leaf list =
   ...


let augment (piece:Piece.t) : (leaf list * () list) =
   let leafs    = flattenWhole piece
   and augments = augmentBlank piece in
   (leafs , augments)


let augment (piece:Piece.t)
   : (leaf list * ((tuning * (pitch * time)) list)) =

   let leafs   = flattenWhole piece

   and tunings = augmentTuning piece
   and pitchs  = augmentPitch  piece
   and times   = augmentTime   piece in

   try
      let pitchsTimes        = List.combine pitchs times in
      let tuningsPitchsTimes = List.combine tunings pitchsTimes in
   catch
      Invalid_argument

   let combineo (lao:a option list) (lbo:b option list) : (a * b option) list =
      let combiner (ao:a option) (bo:b option) =
         longer
         ...
      in
   val unfoldo : ?list:('a list) -> ?index:int -> (int->'a option) -> 'a list
      List.unfold combiner

   (* as longest *)
   let rec combineo ?(out = []) (la:a option list) (lb:b option list)
      : (a * b) list =
      match la , lb with
      | []         , []         -> out
      | []         , bHd :: bTl -> combineo ((None , bHd) :: out) []  bTl
      | aHd :: aTl , []         -> combineo ((aHd , None) :: out) aTl []
      | aHd :: aTl , bHd :: bTl -> combineo ((aHd , bHd)  :: out) aTl bTl

   (leafs , tuningsPitchsTimesHeights)


(* ? the following in Print module: *)

let applyAugments (leaf:leaf) (blank:()) : string =

   | Str        str        -> stringPrint str
   | Comment    line       -> commentPrint line
   | Metadata   metadata   -> metadataPrint metadata
   | Annotation annotation -> annotationPrint annotation
   | Indicator  indicator  -> indicatorPrint indicator
   | Sound      sound      -> soundPrint sound


let applyAugments
   (leaf:leaf) (tuning , (pitch , time) ; tuning * (pitch * time))
   : string =

   match leaf with
   | Str        str        -> stringPrint str
   | Comment    line       -> commentPrint line
   | Metadata   metadata   -> metadataPrint metadata
   | Annotation annotation -> annotationPrint annotation
   | Indicator  indicator  -> indicatorPrint indicator
   | Sound      sound      -> soundPrint sound tuning pitch time


let print (formats:formats) (piece:Piece.t) : string =

   let leafs , augments = Augment.augment piece in
   let strings          = List.map2 applyAugments leafs augments in
   String.concat "" strings

*)



(*
   let augfolder (augmented , tuning : Piece.t * Tuning.t) (item:Piece.item)
      : Piece.t * Tuning.t =
      let itemNew , tuningNew =
         match item with
         | Content    contentList ->
            (* augment content *)
            let augmapper (content:content) : content =
               match content with
               | Sound     sound  -> ...
                  augment the tree with note-pitchs,
                  on each string/fret-note, add pitch,
                  calculated from last tuning
                  (move some stuff out of Print and Pitch (wordToNote))
               | Indicator _ as i -> i
            in
            (List.map augmapper contentList , tuning)
         | Annotation annotation  ->
            (* update current tuning *)
            (item , Tuning.create annotation)
         | Metadata   _
         | Comment    _
         | Blankline    -> (item , tuning)
      in
      (itemNew :: augmented , tuningNew)

   and tuningDefault = Tuning.create Annotation.default in

   let augmented , _ = List.fold_left augfolder tuningDefault piece in
   (* TODO?: maybe reverse list ? *)
   augmented
*)



(*
   let annotationSumject (item:item) : Annotation.t option =
      (* like projection on a product, but for a sum *)
      match item with
      | Annotation a -> Some a
      | Content    _ -> None
      | Metadata   _ -> None
      | Comment    _ -> None
      | Blankline    -> None

   and trailulator (default:'a) (lo:'a option list) : 'a list =
      (* fill gaps in an option list with the last one *)
      let trailFolder (default:'a) (l:'a list) (o:'a option) : 'a list =
         let v =
            match o with
            | Some a -> a
            | None   ->
               begin match l with
               | []        -> default
               | prev :: _ -> prev
               end
         in
         v :: l
      in
      (*List.fold_right (trailFolder default) lo []*)
      (List.fold_left (trailFolder default) [] lo)
      |>
      List.rev

   and augmenter (item , tuning :(item * Tuning.t)) : augitem =
      match item with
      | Content contentList ->
         let augmenter (content:content) : content =
            match content with
            | Sound     sound  -> ...
               begin match sound with
               | Ordinary { _ ; pitch ; _ } as sound ->

Pitch
type t     = { pitch:pitch ; chord:pitch list ; }
type pitch = { basic:basic ; extra:extra ; }
type basic = | Word of word | Note of note
type word  = { strng :strng ; fret  :fret  ; }

                  let augmentPitch (pitch:Pitch) : augPitch =
                     match pitch.basic with
                     | Word { strng ; fret } ->
                        Tuning.note tuning strng fret
                     | Note _ as note -> note

                  augmentPitch pitch
                  List.map augmentPitch chord

                  (sound , somethinger pitch)
               | Continuation _
               | Rest         _ as unchanged -> unchanged
               end
            | Indicator _ as unchanged -> unchanged
         in
         List.map augmenter contentList
      | Annotation _
      | Metadata   _
      | Comment    _
      | Blankline as unchanged -> unchanged
   in

   (* make a sparse list : (Annotation.t option) list
      (_ 1 _ _ _ 2 _ 3 _ _) *)
   (List.map annotationSumject piece)
   |>
   (* fill gaps with last value : Annotation.t list
      (0 1 1 1 1 2 2 3 3 3) *)
   (trailulator Annotation.default)
   |>
   (* extract tuning : Tuning.t list *)
   (List.map Tuning.create)
   |>
   (* zip with original item list : (item * Tuning.t) list *)
   (List.combine piece)
   |>
   (* augment elements *)
   (List.map augmenter)
*)
